package com.nisum.BookInfo.service;

import com.nisum.BookInfo.dao.BookDao;
import com.nisum.BookInfo.dto.BookDto;
import com.nisum.BookInfo.entity.Book;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.springframework.boot.test.context.SpringBootTest;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;
import reactor.test.StepVerifier;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.*;

@SpringBootTest

class BookServiceImplTest {

    @Mock
    private BookDao bookDao;  // Mock the BookDao

    @InjectMocks
    private BookServiceImpl bookService;  // Service under test

    private BookDto bookDto;
    private Book book;

    @BeforeEach
    public void setUp() {
        // Set up test data
        bookDto = new BookDto(1, "C++", "This is c++", "C publiser", "Test Author");
        book = new Book(1, "C++", "This is c++", "C publiser", "Test Author");


        //when(bookDao.saveBook(any(Book.class))).thenReturn(Mono.just(book));
        MockitoAnnotations.openMocks(this);

    }

    @Test
    public void saveBook_shouldSaveBookAndReturnMonoBook() {
        // Call the saveBook method of the service
        when(bookDao.saveBook(any(Book.class))).thenReturn(Mono.just(book));
        Mono<Book> result = bookService.saveBook(bookDto);

        // Verify the result using StepVerifier (for reactive streams)
        StepVerifier.create(result)
                .expectNextMatches(savedBook -> {
                    // Assert the values in the saved Book are as expected
                    assertEquals("C++", savedBook.getName());
                    assertEquals("Test Author", savedBook.getAuthor());
                    assertEquals("C publiser", savedBook.getPubliser());
                    return true;
                })
                .verifyComplete();

        // Verify that bookDao.saveBook was called exactly once
        verify(bookDao, times(1)).saveBook(any(Book.class));
    }

    @Test
    public void getAllBooks_shouldReturnFluxOfBooks() {
        // Prepare a Flux of books for testing
        Book book1 = new Book(1, "C++", "This is c++", "C publiser", "Test Author");
        Book book2 = new Book(2, "Java", "This is c++", "C publiser", "Test Author");
        Flux<Book> bookFlux = Flux.just(book1, book2);

        // Mock the getAllBooks method to return the Flux of books
        when(bookDao.getAllBooks()).thenReturn(bookFlux);
        Flux<Book> result = bookService.getAllBooks();

        // Verify the result using StepVerifier (for reactive streams)
        StepVerifier.create(result)
                .expectNextMatches(book -> book.getName().equals("C++"))
                .expectNextMatches(book -> book.getName().equals("Java"))
                .verifyComplete();

        // Verify that bookDao.getAllBooks was called exactly once
        verify(bookDao, times(1)).getAllBooks();
    }

    @Test
    public void getBookById_shouldReturnBook_whenBookExists() {
        //Setup
        int bookId = 1;
        book = new Book(bookId, "C++", "This is c++", "C publiser", "Test Author");

        // Mock the getBookById method to return the expected Mono<Book>
        when(bookDao.getBookById(bookId)).thenReturn(Mono.just(book));
        //setupend

        // Call the getBookById method of the service
        Mono<Book> result = bookService.getBookById(bookId);

        // Verify the result using StepVerifier (for reactive streams)
        StepVerifier.create(result)
                .expectNextMatches(savedBook -> {
                    // Assert the book details
                    assertEquals(bookId, savedBook.getBookId());
                    assertEquals("C++", savedBook.getName());
                    assertEquals("Test Author", savedBook.getAuthor());
                    assertEquals("C publiser", savedBook.getPubliser());
                    return true;
                })
                .verifyComplete();

        // Verify that bookDao.getBookById was called exactly once
        verify(bookDao, times(1)).getBookById(bookId);
    }

    @Test
    public void getBookById_shouldReturnEmpty_whenBookDoesNotExist() {
        //setup
        int bookId = 1;
        book = new Book(bookId, "C++", "This is c++", "C publiser", "Test Author");

        // Mock the getBookById method to return the expected Mono<Book>
        when(bookDao.getBookById(bookId)).thenReturn(Mono.just(book));
        //setupend
        // Mock the DAO method to return an empty Mono when no book is found
        when(bookDao.getBookById(999)).thenReturn(Mono.empty());

        // Call the getBookById method with a non-existent ID
        Mono<Book> result = bookService.getBookById(999);

        // Verify the result is empty (no book found)
        StepVerifier.create(result)
                .expectComplete()  // Should complete without emitting anything
                .verify();

        // Verify that bookDao.getBookById was called exactly once
        verify(bookDao, times(1)).getBookById(999);
    }


    @Test
    public void updateBook_shouldReturnUpdatedBook_whenBookExists() {
        int bookId = 1;
        bookDto = new BookDto(1, "C++", "This is c++", "C publiser", "Test Author");
        book = new Book(1, "C++", "This is c++", "C publiser", "Test Author");
        // Mock the DAO method to return the updated book
        when(bookDao.updateBook(eq(bookId), any(Book.class))).thenReturn(Mono.just(book));

        // Call the updateBook method of the service
        Mono<Book> result = bookService.updateBook(bookId, bookDto);

        // Verify the result using StepVerifier
        StepVerifier.create(result)
                .expectNextMatches(updatedBook -> {
                    // Assert that the book is updated with the correct fields
                    assertEquals(bookId, updatedBook.getBookId());
                    assertEquals("C++", updatedBook.getName());
                    assertEquals("Test Author", updatedBook.getAuthor());
                    assertEquals("C publiser", updatedBook.getPubliser());
                    return true;
                })
                .verifyComplete();

        // Verify that bookDao.updateBook was called with the correct bookId and book
        verify(bookDao, times(1)).updateBook(eq(bookId), any(Book.class));
    }

    @Test
    public void updateBook_shouldReturnEmpty_whenBookDoesNotExist() {
        int bookId = 1;
        bookDto = new BookDto(1, "C++", "This is c++", "C publiser", "Test Author");
        book = new Book(1, "C++", "This is c++", "C publiser", "Test Author");
        // Mock the DAO method to return an empty Mono (book not found)
        when(bookDao.updateBook(eq(999), any(Book.class))).thenReturn(Mono.empty());

        // Call the updateBook method with a non-existent bookId
        Mono<Book> result = bookService.updateBook(999, bookDto);

        // Verify that the result is empty (no book found to update)
        StepVerifier.create(result)
                .expectComplete()  // Should complete without emitting anything
                .verify();

        // Verify that bookDao.updateBook was called with the non-existent bookId
        verify(bookDao, times(1)).updateBook(eq(999), any(Book.class));
    }

    @Test
    public void deleteBook_shouldCompleteSuccessfully_whenBookExists() {
        int  bookId = 1;
        // Mock the DAO method to simulate successful deletion
        when(bookDao.deleteBook(bookId)).thenReturn(Mono.empty());

        // Call the deleteBook method of the service
        Mono<Void> result = bookService.deleteBook(bookId);

        // Verify the result using StepVerifier
        StepVerifier.create(result)
                .verifyComplete();  // Should complete without emitting anything

        // Verify that bookDao.deleteBook was called exactly once with the correct bookId
        verify(bookDao, times(1)).deleteBook(bookId);
    }

    @Test
    public void deleteBook_shouldCompleteSuccessfully_whenBookDoesNotExist() {
        int  bookId = 1;
        // Mock the DAO method to simulate a case where no book is found
        when(bookDao.deleteBook(999)).thenReturn(Mono.empty());

        // Call the deleteBook method with a non-existent bookId
        Mono<Void> result = bookService.deleteBook(999);

        // Verify that the result completes without emitting any value (book not found)
        StepVerifier.create(result)
                .verifyComplete();  // Should complete without emitting anything

        // Verify that bookDao.deleteBook was called exactly once with the non-existent bookId
        verify(bookDao, times(1)).deleteBook(999);
    }

        }

